"""FERC 1 specific transformation parameters.

These constants are used to construct :mod:`pydantic` models, which are validated and
used to control the various data transformations. The definitions of those models can be
found in :mod:`pudl.transform.classes` and :mod:`pudl.transform.ferc1`
"""

from datetime import date

import numpy as np
import pandas as pd

##############################################################################
# Unit converstion parameters
##############################################################################
PERPOUND_TO_PERSHORTTON = dict(
    multiplier=2000.0,
    from_unit=r"_per_lb",
    to_unit=r"_per_ton",
)
"""Parameters for converting from inverse pounds to inverse short tons."""

CENTS_TO_DOLLARS = dict(
    multiplier=0.01,
    from_unit=r"_cents",
    to_unit=r"_usd",
)
"""Parameters for converting from cents to dollars."""

CENTS_PERMMBTU_TO_USD_PERMMBTU = dict(
    multiplier=0.01,
)
"""Parameters for converting from cents per mmbtu to dollars per mmbtu."""


PERCF_TO_PERMCF = dict(
    multiplier=1000.0,
    from_unit=r"_per_cf",
    to_unit=r"_per_mcf",
)
"""Parameters for converting from inverse cubic feet to inverse 1000s of cubic feet."""

PERGALLON_TO_PERBARREL = dict(
    multiplier=42.0,
    from_unit=r"_per_gal",
    to_unit=r"_per_bbl",
)
"""Parameters for converting from inverse gallons to inverse barrels."""

PERKW_TO_PERMW = dict(
    multiplier=1000.0,
    from_unit=r"_per_kw",
    to_unit=r"_per_mw",
)
"""Parameters for converting column units from per kW to per MW."""

PERKWH_TO_PERMWH = dict(
    multiplier=1000.0,
    from_unit=r"_per_kwh",
    to_unit=r"_per_mwh",
)
"""Parameters for converting column units from per kWh to per MWh."""

KW_TO_MW = dict(
    multiplier=1e-3,
    from_unit=r"_kw",
    to_unit=r"_mw",
)
"""Parameters for converting column units from kW to MW."""

KWH_TO_MWH = dict(
    multiplier=1e-3,
    from_unit=r"_kwh",
    to_unit=r"_mwh",
)
"""Parameters for converting column units from kWh to MWh."""

BTU_TO_MMBTU = dict(
    multiplier=1e-6,
    from_unit=r"_btu",
    to_unit=r"_mmbtu",
)
"""Parameters for converting column units from BTU to MMBTU."""

PERBTU_TO_PERMMBTU = dict(
    multiplier=1e6,
    from_unit=r"_per_btu",
    to_unit=r"_per_mmbtu",
)
"""Parameters for converting column units from BTU to MMBTU."""

BTU_PERKWH_TO_MMBTU_PERMWH = dict(
    multiplier=(1e-6 * 1000.0),
    from_unit=r"_btu_per_kwh",
    to_unit=r"_mmbtu_per_mwh",
)
"""Parameters for converting column units from BTU/kWh to MMBTU/MWh."""


##############################################################################
# Valid ranges to impose on various columns
##############################################################################
VALID_PLANT_YEARS = {
    "lower_bound": 1850,
    "upper_bound": date.today().year,
}
"""Valid range of years for power plant construction."""

VALID_COAL_MMBTU_PER_TON = {
    "lower_bound": 6.5,
    "upper_bound": 29.0,
}
"""Valid range for coal heat content, taken from the EIA-923 instructions.

Lower bound is for waste coal. Upper bound is for bituminous coal.

https://www.eia.gov/survey/form/eia_923/instructions.pdf
"""

VALID_COAL_USD_PER_MMBTU = {
    "lower_bound": 0.5,
    "upper_bound": 7.5,
}
"""Historical coal price range from the EIA-923 Fuel Receipts and Costs table."""

VALID_GAS_MMBTU_PER_MCF = {
    "lower_bound": 0.3,
    "upper_bound": 3.3,
}
"""Valid range for gaseous fuel heat content, taken from the EIA-923 instructions.

Lower bound is for landfill gas. Upper bound is for "other gas".  Blast furnace gas
(which has very low heat content) is effectively excluded.

https://www.eia.gov/survey/form/eia_923/instructions.pdf
"""

VALID_GAS_USD_PER_MMBTU = {
    "lower_bound": 1.0,
    "upper_bound": 35.0,
}
"""Historical natural gas price range from the EIA-923 Fuel Receipts and Costs table."""

VALID_OIL_MMBTU_PER_BBL = {
    "lower_bound": 3.0,
    "upper_bound": 6.9,
}
"""Valid range for petroleum fuels heat content, taken from the EIA-923 instructions.

Lower bound is for waste oil. Upper bound is for residual fuel oil.

https://www.eia.gov/survey/form/eia_923/instructions.pdf
"""

VALID_OIL_USD_PER_MMBTU = {
    "lower_bound": 5.0,
    "upper_bound": 33.0,
}
"""Historical petroleum price range from the EIA-923 Fuel Receipts and Costs table."""

##############################################################################
# Unit Corrections
##############################################################################
COAL_COST_PER_MMBTU_CORRECTIONS = {
    "data_col": "fuel_cost_per_mmbtu",
    "cat_col": "fuel_type_code_pudl",
    "cat_val": "coal",
    "valid_range": VALID_COAL_USD_PER_MMBTU,
    "unit_conversions": [
        CENTS_TO_DOLLARS,
    ],
}
GAS_COST_PER_MMBTU_CORRECTIONS = {
    "data_col": "fuel_cost_per_mmbtu",
    "cat_col": "fuel_type_code_pudl",
    "cat_val": "gas",
    "valid_range": VALID_GAS_USD_PER_MMBTU,
    "unit_conversions": [
        CENTS_TO_DOLLARS,
    ],
}
OIL_COST_PER_MMBTU_CORRECTIONS = {
    "data_col": "fuel_cost_per_mmbtu",
    "cat_col": "fuel_type_code_pudl",
    "cat_val": "oil",
    "valid_range": VALID_OIL_USD_PER_MMBTU,
    "unit_conversions": [
        CENTS_TO_DOLLARS,
    ],
}
FUEL_COST_PER_MMBTU_CORRECTIONS = [
    COAL_COST_PER_MMBTU_CORRECTIONS,
    GAS_COST_PER_MMBTU_CORRECTIONS,
    OIL_COST_PER_MMBTU_CORRECTIONS,
]

COAL_MMBTU_PER_UNIT_CORRECTIONS = {
    "data_col": "fuel_mmbtu_per_unit",
    "cat_col": "fuel_type_code_pudl",
    "cat_val": "coal",
    "valid_range": VALID_COAL_MMBTU_PER_TON,
    "unit_conversions": [
        PERPOUND_TO_PERSHORTTON,
        BTU_TO_MMBTU,
    ],
}
GAS_MMBTU_PER_UNIT_CORRECTIONS = {
    "data_col": "fuel_mmbtu_per_unit",
    "cat_col": "fuel_type_code_pudl",
    "cat_val": "gas",
    "valid_range": VALID_GAS_MMBTU_PER_MCF,
    "unit_conversions": [
        PERCF_TO_PERMCF,
        BTU_TO_MMBTU,
    ],
}
OIL_MMBTU_PER_UNIT_CORRECTIONS = {
    "data_col": "fuel_mmbtu_per_unit",
    "cat_col": "fuel_type_code_pudl",
    "cat_val": "oil",
    "valid_range": VALID_OIL_MMBTU_PER_BBL,
    "unit_conversions": [
        PERGALLON_TO_PERBARREL,
        BTU_TO_MMBTU,  # Why was this omitted in the old corrections?
    ],
}
FUEL_MMBTU_PER_UNIT_CORRECTIONS = [
    COAL_MMBTU_PER_UNIT_CORRECTIONS,
    GAS_MMBTU_PER_UNIT_CORRECTIONS,
    OIL_MMBTU_PER_UNIT_CORRECTIONS,
]

##############################################################################
# String normalizations
##############################################################################
FERC1_STRING_NORM = {
    "remove_chars": r"$?={}\x02\x00",
    "nullable": False,
}

##############################################################################
# String categorizations
##############################################################################
FUEL_CATEGORIES: dict[str, set[str]] = {
    "categories": {
        "coal": {
            "bit coal",
            "ciak",
            "coa",
            "coal",
            "coal & gas",
            "coal & oil",
            "coal (sb)",
            "coal - subbit",
            "coal and gas",
            "coal bit",
            "coal mcf",
            "coal tons",
            "coal unit #3",
            "coal&coke",
            "coal(sb)",
            "coal-lig",
            "coal-lignite",
            "coal-sub",
            "coal-sub bit",
            "coal-subb",
            "coal-subbit",
            "coal-subbitum",
            "coal-u3",
            "coal-unit #3",
            "coal. gas",
            "coal.oil",
            "coal/gas",
            "coal/oil",
            "coals",
            "coke",
            "fossil",
            "lignite",
            "lignite/coal",
            "other-steam expenses:",
            "pet. coke",
            "petcoke",
            "steam plants",
            "steam-common production",
            "steam-common production plant",
            "tons",
        },
        "oil": {
            "# 2  oil",
            "# 2 oil",
            "# 6 oil",
            "#1 diesel",
            "#1 uls diesel",
            "#1 fuel oil",
            "#1 kerosene",
            "#1 ls diesel",
            "#1 oil",
            "#2",
            "#2 diesel",
            "#2 fue oil",
            "#2 fuel",
            "#2 fuel oil",
            "#2 ls diesel",
            "#2 uls diesel",
            "#2 oil",
            "#2 oil-diesel",
            "#2oil diesel",
            "#6",
            "#6 & used",
            "#6 hvy oil",
            "#6 oil",
            "2",
            "2 oil",
            "6",
            "barrels",
            "bbl",
            "bbl.",
            "biofuel",
            "black start diesel units",
            "black start diesel units:",
            "blend oil",
            "blended oil",
            "bunker oil",
            "comp oil/gas",
            "dielel oil",
            "dielsel oil",
            "diesal",
            "diesel",
            "diesel fuel",
            "diesel generators",
            "diesel oil",
            "diesel plant",
            "diesel plants",
            "diesel plants:",
            "diesel:",
            "diesel-low",
            "diesel-low s",
            "diesel-low s or gas",
            "diesel-ultra low s",
            "diesel/",
            "diesel/compo",
            "diesel/compos",
            "diesle-low s",
            "fuel oil",
            "fo",
            "heavy oil",
            "jet",
            "jet fuel",
            "jet oil",
            "kero",
            "light oil",
            "montauk diesel",
            "nno. 2 oil",
            "no 2",
            "no. 2 diesel",
            "no 2 oil",
            "no 2. oil",
            "no 6 oil",
            "no. 2",
            "no. 2  oil",
            "no. 2 oil",
            "no. 2 oils",
            "no. 6 oil",
            "no.2 oil",
            "no.2 oil gas",
            "no.6 oil",
            "no.6& used",
            "oial",
            "oil",
            "oil #6",
            "oil (#2)",
            "oil (#6)",
            "oil (6)",
            "oil (used)",
            "oil - 2",
            "oil / gas",
            "oil 6",
            "oil bbl",
            "oil bbls",
            "oil bls",
            "oil un 1 & 2",
            "oil(#2)",
            "oil-2",
            "oil-6",
            "oil-8",
            "oil-diesel",
            "oil-fuel",
            "oil-gas",
            "oil-kero",
            "oil-u1&2",
            "oil-unit #1",
            "oil-unit1&2",
            "oil.",
            "oil, nat .gas",
            "oil, nat. gas",
            "oil,n.gas",
            "oil,nat.gas",
            "oil/diesel",
            "oil/gas",
            "oill",
            "oils",
            "oiul",
            "oli",
            "pil",
            "resid. oil",
            "residual oil",
            "tall oil",
            "used oil",
        },
        "gas": {
            "# 6 gas",
            "#6 gas",
            "* propane",
            "blast gaa",
            "blast gas",
            "coal.gas",
            "coke gas",
            "coke oven gas",
            "dinner lake gas",
            "g",
            "ga",
            "gac",
            "gas",
            "gas #2016",
            "gas & oil",
            "gas - oil",
            "gas / oil",
            "gas ct",
            "gas expander",
            "gas mcf",
            "gas plant",
            "gas turbine",
            "gas turbine plant",
            "gas turbine plants",
            "gas turbine:",
            "gas turbines",
            "gas turbines:",
            "gas, fuel oil",
            "gas, oil",
            "gas-oil",
            "gas.oil",
            "gas/oil",
            "gas/fuel oil",
            "gas`",
            "gass",
            "lp gas",
            "mcf",
            "methane",
            "methane gas",
            "n. gas",
            "nat  gas",
            "nat gas",
            "nat. gas",
            "nat.gas",
            "natl gas",
            "natuaral gas",
            "natural",
            "natural  gas",
            "natural gas",
            "ng",
            "ng, fo",
            "prop",
            "prop.",
            "propane",
            "propane **",
            "propane**",
            "syngas",
            "syngass",
            "waste heat/gas",
        },
        "solar": {
            "community solar gardens",
            "photovoltaic",
            "photo voltaic generating plants:",
            "proj dvlpmnt div solar photovoltaic project:",
            "sixth street solar",
            "solar",
            "solar array",
            "solar electric",
            "solar photovoltaic",
            "solar photovoltiac",
            "solar plant",
            "solar plants",
            "solar plants:",
            "solar:",
            "other production - solar",
        },
        "wind": {
            "rosiere wind turbines",
            "wind",
            "wind turbine",
            "wind turbine plant",
            "wind turbine plants",
            "wind turbine plants:",
            "wind turbine:",
            "wind turbines",
            "wind:",
        },
        "hydro": {
            "additional hydro plant",
            "conventional hydro",
            "hydraulic",
            "hydraulic (1):",
            "hydraulic:",
            "hydro",
            "hyrdo",
            "hydro -",
            "hydro : licensed proj. no.",
            "hydro electric",
            "hydro lic project no.",
            "hydro lic. project no.",
            "hydro license",
            "hydro license no.",
            "hydro plant:",
            "hydro plants",
            "hydro plants:",
            "hydro plants: licensed proj no.",
            "hydro plants: licensed proj. no.",
            "hydro-electric",
            "hydro-neal shoals",
            "hydro:",
            "hydro: lic project no.",
            "hydro: lic. project no.",
            "hydro: license no.",
            "hydroelectric",
            "hydroelectric : licensed proj. no.",
            "hydroelectric generating plants",
            "hydroelectric generating plants :",
            "hydroelectric generating plants:",
            "hydroelectric: licensed proj. no.",
            "hyrdo",
            "other hydro assets (resvr., strm gages, etc.)",
            "other hydro assets (resvr.,strm gages, etc.)",
            "other hydro costs:",
            "research and development costs (hydros)",
            "run of river",
            "run-of-river",
            "run-of-rr",
            "run-of-rv",
            "small hydroelectric plants:",
            "small hydroelectric generating plants:",
            "unallocated hydros",
            "unallocated - hydros",
            "water",
        },
        "nuclear": {
            "nuclear",
            "grams of uran",
            "grams of",
            "grams of  ura",
            "grams",
            "nucleur",
            "nulear",
            "nucl",
            "nucleart",
            "nucelar",
            "gr.uranium",
            "grams of urm",
            "nuclear (9)",
            "nulcear",
            "nuc",
            "gr. uranium",
            "uranium",
            "nuclear mw da",
            "grams of ura",
            "nucvlear",
            "nuclear (1)",
        },
        "waste": {
            "biomass",
            "landfill",
            "landfill gas",
            "landfill:",
            "rdf",
            "rdf, gas",
            "refuse",
            "switchgrass",
            "tdf",
            "tire",
            "tire chips",
            "tire refuse",
            "tires",
            "tires/refuse",
            "waste",
            "waste oil",
            "wood",
            "wood chips",
            "wood waste",
            "woodchips",
            "woodships",
        },
        "other": {  # This should really be NA but we require a fuel_type_code_pudl
            "",
            "# 6 & used",
            "*",
            "/#2a",
            "0",
            "1",
            "120976",
            "20",
            "3",
            "77509",
            "a",
            "a, b",
            "a,b",
            "all",
            "all fuel",
            "all fuels",
            "all fuels =",
            "all fuels are",
            "all fuels to",
            "alt. fuel",
            "alt. fuels",
            "alternative",
            "at right",
            "average",
            "avg",
            "avg.",
            "bio fuel",
            "blo",
            "blo bls",
            "bng",
            "c",
            "c omposite",
            "comb",
            "comb.",
            "combined",
            "comp",
            "compos.",
            "composit",
            "composite",
            "compsite",
            "fuel cell:",
            "fuel cell",
            "furfural",
            "kagv",
            "lime",
            "m",
            "mbtu's",
            "mbtus",
            "mw days",
            "n.a.",
            "n/a",
            "na",
            "none",
            "other",
            "purch steam",
            "purch. steam",
            "rrm",
            "steam",
            "the right are",
            "toal",
            "total",
            "total pc",
            "total pr crk",
            "total prairie",
            "waste heat",
            "watson bng",
            "—",
            "-",
            "--",
            "---------",
            '"',
        },
        "na_category": {
            "na_category",
            "internal combustion",
            "internal combustion:",
            "internal combustion plants:",
            "gold creek internal combustion:",
        },
    }
}
"""
A mapping a canonical fuel name to a set of strings which are used to represent that
fuel in the FERC Form 1 Reporting. Case is ignored, as all fuel strings are converted to
lower case in the data set.
"""

FUEL_UNIT_CATEGORIES: dict[str, set[str]] = {
    "categories": {
        "ton": {
            "ton",
            "toms",
            "taons",
            "tones",
            "col-tons",
            "toncoaleq",
            "coal",
            "tons coal eq",
            "coal-tons",
            "tons",
            "tons coal",
            "coal-ton",
            "tires-tons",
            "coal tons -2 ",
            "oil-tons",
            "coal tons 200",
            "ton-2000",
            "coal tons",
            "coal tons -2",
            "coal-tone",
            "tire-ton",
            "tire-tons",
            "ton coal eqv",
            "tos",
            "coal tons - 2",
            "c. t.",
            "c.t.",
            "t",
            "toncoalequiv",
        },
        "mcf": {
            "mcf",
            "mcf's",
            "mcfs",
            "mcf.",
            "mcfe",
            "gas mcf",
            '"gas" mcf',
            "gas-mcf",
            "mfc",
            "mct",
            " mcf",
            "msfs",
            "mlf",
            "mscf",
            "mci",
            "mcl",
            "mcg",
            "m.cu.ft.",
            "kcf",
            "(mcf)",
            "mcf *(4)",
            "mcf00",
            "m.cu.ft..",
            "1000 c.f",
        },
        "bbl": {
            "bbl",
            "barrel",
            "bbls",
            "barrels",
            "bbrl",
            "bbl.",
            "bbls.",
            "oil 42 gal",
            "oil-barrels",
            "barrrels",
            "bbl-42 gal",
            "oil-barrel",
            "bb.",
            "barrells",
            "bar",
            "bbld",
            "oil- barrel",
            "barrels    .",
            "bbl .",
            "barels",
            "barrell",
            "berrels",
            "bb",
            "bbl.s",
            "oil-bbl",
            "bls",
            "bbl:",
            "barrles",
            "blb",
            "propane-bbl",
            "barriel",
            "berriel",
            "barrile",
            "(bbl.)",
            "barrel *(4)",
            "(4) barrel",
            "bbf",
            "blb.",
            "(bbl)",
            "bb1",
            "bbsl",
            "barrrel",
            "barrels 100%",
            "bsrrels",
            "bbl's",
            "*barrels",
            "oil - barrels",
            "oil 42 gal ba",
            "bll",
            "boiler barrel",
            "gas barrel",
            '"boiler" barr',
            '"gas" barrel',
            '"boiler"barre',
            '"boiler barre',
            "barrels .",
            "bariel",
            "brrels",
            "oil barrel",
            "barreks",
            "oil-bbls",
            "oil-bbs",
            "boe",
        },
        "mmbbl": {"mmbbl", "mmbbls"},
        "gal": {"gal", "gallons", "gal.", "gals", "gals.", "gallon", "galllons"},
        "kgal": {
            "kgal",
            "oil(1000 gal)",
            "oil(1000)",
            "oil (1000)",
            "oil(1000",
            "oil(1000ga)",
            "1000 gals",
            "1000 gal",
        },
        "grams": {
            "gram",
            "grams",
            "gm u",
            "grams u235",
            "grams u-235",
            "grams of uran",
            "grams: u-235",
            "grams:u-235",
            "grams:u235",
            "grams u308",
            "grams: u235",
            "grams of",
            "grams - n/a",
            "gms uran",
            "s e uo2 grams",
            "gms uranium",
            "grams of urm",
            "gms. of uran",
            "grams (100%)",
            "grams v-235",
            "se uo2 grams",
            "grams u",
            "g",
            "grams of uranium",
        },
        "kg": {
            "kg",
            "kg of uranium",
            "kg uranium",
            "kilg. u-235",
            "kg u-235",
            "kilograms-u23",
            "kilograms u-2",
            "kilograms",
            "kg of",
            "kg-u-235",
            "kilgrams",
            "kilogr. u235",
            "uranium kg",
            "kg uranium25",
            "kilogr. u-235",
            "kg uranium 25",
            "kilgr. u-235",
            "kguranium 25",
            "kg-u235",
            "kgm",
        },
        "klbs": {
            "klbs",
            "k lbs.",
            "k lbs",
            "1000 / lbs",
            "1000 lbs",
        },
        "mmbtu": {
            "mmbtu",
            "mmbtus",
            "mbtus",
            "(mmbtu)",
            "mmbtu's",
            "nuclear-mmbtu",
            "nuclear-mmbt",
            "mmbtul",
        },
        "btu": {
            "btu",
            "btus",
        },
        "mwdth": {
            "mwdth",
            "mwd therman",
            "mw days-therm",
            "mwd thrml",
            "mwd thermal",
            "mwd/mtu",
            "mw days",
            "mwd",
            "mw day",
            "dth",
            "mwdaysthermal",
            "mw day therml",
            "mw days thrml",
            "nuclear mwd",
            "mmwd",
            "mw day/therml" "mw days/therm",
            "mw days (th",
            "ermal)",
        },
        "mwhth": {
            "nwh therm",
            "mwhth",
            "mwh them",
            "mwh threm",
            "mwh therm",
            "mwh",
            "mwh therms.",
            "mwh term.uts",
            "mwh thermal",
            "mwh thermals",
            "mw hr therm",
            "mwh therma",
            "mwh therm.uts",
        },
        "na_category": {
            "na_category",
            "na",
            "",
            "1265",
            "mwh units",
            "composite",
            "therms",
            "n/a",
            "mbtu/kg",
            "uranium 235",
            "oil",
            "ccf",
            "2261",
            "uo2",
            "(7)",
            "oil #2",
            "oil #6",
            'a"',
            "dekatherm",
            "0",
            "mw day/therml",
            "nuclear",
            "gas",
            "62,679",
            "mw days/therm",
            "uranium",
            "oil/gas",
            "thermal",
            "(thermal)",
            "se uo2",
            "181679",
            "83",
            "3070",
            "248",
            "273976",
            "747",
            "-",
            "are total",
            "pr. creek",
            "decatherms",
            "uramium",
            ".",
            "total pr crk",
            ">>>>>>>>",
            "all",
            "total",
            "alternative-t",
            "oil-mcf",
            "3303671",
            "929",
            "7182175",
            "319",
            "1490442",
            "10881",
            "1363663",
            "7171",
            "1726497",
            "4783",
            "7800",
            "12559",
            "2398",
            "creek fuels",
            "propane-barre",
            "509",
            "barrels/mcf",
            "propane-bar",
            "4853325",
            "4069628",
            "1431536",
            "708903",
            "mcf/oil (1000",
            "344",
            "mcf / gallen",
            "none",
            "—",
        },
    }
}
"""
A mapping of canonical fuel units (keys) to sets of strings representing those
fuel units (values)
"""

PLANT_TYPE_CATEGORIES: dict[str, set[str]] = {
    "categories": {
        "steam": {
            "*resp. share steam",
            "(6) milwaukee county is a steam utility plant",
            "(steam)",
            "40% share steam",
            "40% share steamsteam (2)",
            "coal",
            "coal fired steam tur",
            "coal fired steam turbine",
            "coal-fired steam",
            "mpc 50% share steam",
            "mpc 50% shares steam",
            "mpc 50%share steam",
            "mpc40%share steam",
            "mpc50% share steam",
            "mpc50% share steam",
            "oil-fired steam",
            "other-steam expenses:",
            "resp share st note 3",
            "resp share st note3",
            "resp share stm note3",
            "resp share stm note3",
            "resp. share steam",
            "respondent's sh-st",
            "respondent's sh.-st.",
            "steam",
            "steam (1)",
            "steam (1) & (2)",
            "steam (2)",
            "steam (3)",
            "steam (4)",
            "steam (5)",
            "steam (6)",
            "steam (7)",
            "steam (8)",
            "steam (a)",
            "steam (a,b)",
            "steam (incl i.c.)",
            "steam (leased)",
            "steam (note 1)",
            "steam (retired)",
            "steam (see note 1)",
            "steam (see note 1,",
            "steam (see note 3)",
            "steam (see notes)",
            "steam - 100%",
            "steam - 64%",
            "steam - 72%",
            "steam 1",
            "steam 100",
            "steam 4, 5",
            "steam a",
            "steam conventional",
            "steam fossil",
            "steam heating plant",
            "steam plant",
            "steam plants",
            "steam retired 2013",
            "steam turbine",
            "steam units 1 2 3",
            "steam units 1 2 3",
            "steam units 1 and 2",
            "steam units 1, 2, 3",
            "steam units 1,2,3",
            "steam units 3 and 4",
            "steam units 4 & 6",
            "steam units 4 5",
            "steam units 4&5",
            "steam units 4&6",
            "steam units 4, 5",
            "steam- 64%",
            "steam- 72%",
            "steam-100%",
            "steam-64%",
            "steam-common production",
            "steam-common production plant",
            "steam-internal comb",
            "steam/fossil",
            "steam;retired - 2013",
            "steam:",
            "steams",
            "stean",
            "stream",
            "unit total steam",
            "unit total-steam",
            "respondent's sh.-st.",
            "respondent's sh-st",
            "40% share steam",
            "resp share stm note3",
            "mpc50% share steam",
            "resp share st note 3",
            "coal fired steam tur",
            "coal fired steam turbine",
            "steam- 64%",
            "steam/fuel oil",
        },
        "combustion_turbine": {
            "* gas turbine",
            "*gas turbine",
            "*gas turbine(note 3)",
            "*gas turbine(note3)",
            "add to gas turbine",
            "b-combustion turbine unit",
            "com turbine peaking",
            "comb turb peak units",
            "comb turb peaking",
            "comb turbine peaking",
            "comb. turb-gas oper",
            "comb. turb-gas oper.",
            "comb. turb.",
            "comb. turb. oil oper",
            "comb. turbine (a)",
            "comb. turine",
            "comb.turb-gas oper.",
            "comb.turb.gas/oil op",
            "comb.turb.oil oper",
            "comb.turb.oil oper.",
            "comb.turb.peak.units",
            "combusion turbine",
            "combusition turbine",
            "combustine turbine",
            "combustine turbine",
            "combustion and gas",
            "combustion burbine",
            "combustion tubine",
            "combustion turbin",
            "combustion turbine",
            "combustion turbine",
            "combustion turbine",
            "combustion_turbine",
            "combustionturbine",
            "combution turbin",
            "combution turbine",
            "combutsion turbine",
            "comubstion turbine",
            "conbustion turbine",
            "cumbustion turbine",
            "gas - turbine",
            "gas expander turbine",
            "gas plant",
            "gas trubine",
            "gas turb",
            "gas turb(see notes)",
            "gas turb, diesel",
            "gas turb, int. comb",
            "gas turb, int. comb.",
            "gas turb. (see note",
            "gas turb.(see note5)",
            "gas turb.(see notes)",
            "gas turbin (leasd)",
            "gas turbin (lease",
            "gas turbine",
            "gas turbine",
            "gas turbine",
            "gas turbine # 1",
            "gas turbine #1",
            "gas turbine (2)",
            "gas turbine (a)",
            "gas turbine (leasd)",
            "gas turbine (lease",
            "gas turbine (lease)",
            "gas turbine (leased",
            "gas turbine (leased)",
            "gas turbine (note 1)",
            "gas turbine (note 3)",
            "gas turbine (note1)",
            "gas turbine (see nos",
            "gas turbine - note 1",
            "gas turbine note 3,6",
            "gas turbine note 4,6",
            "gas turbine peaker",
            "gas turbine peaking",
            "gas turbine peakload",
            "gas turbine plant",
            "gas turbine plants",
            "gas turbine-simple",
            "gas turbine-unit 2",
            "gas turbine/int comb",
            "gas turbine/int. cm",
            "gas turbine/intcomb",
            "gas turbine:",
            "gas turbine; retired",
            "gas turbines",
            "gas turbines:",
            "gas turbinint comb",
            "gas turbne/int comb",
            "gas-turbine",
            "gasturbine",
            "gt",
            "gtg/gas",
            "i.c.e. /gas turbine",
            "i.c.e./gas tubine",
            "i.c.e./gas turbine",
            "i.c.e/ gas turbine",
            "i.c.e/gas turbine",
            "jet",
            "jet engine",
            "jet powered turbine",
            "jet turbine",
            "resp share gas note3",
            "simple cycle",
            "simple cycle",
            "simple cycle turbine",
            "gas turbine; retired",
            "gas turbine/fuel oil",
        },
        "combined_cycle": {
            "Combined cycle",
            "cc",
            "cc / gas turb",
            "com cycle gas turb",
            "com. cyc",
            "com. cycle",
            "comb cyc",
            "comb cyc (a) (b)",
            "comb cycle",
            "comb cycle (a,b)",
            "comb cycle gas turb",
            "comb cycle/steam tur",
            "comb. cycle",
            "comb. cycle gas turb",
            "combine cycle",
            "combine cycle oper",
            "combined",
            "combined cycle",
            "combined cycle",
            "combined cycle (a,b)",
            "combined cycle - 40%",
            "combined cycle ctg",
            "combined cycle oper",
            "combined cycle opern",
            "combined cycle plant:",
            "combined turbine",
            "combined_cycle",
            "combinedcycle",
            "ctg steam gas",
            "ctg/steam -gas",
            "ctg/steam-gas",
            "gas & steam turbine",
            "gas / steam",
            "gas fired cc turbine",
            "gas steam",
            "gas trb & heat rec",
            "gas tur & comb cyc",
            "gas tur & comb cycl",
            "gas turb-combined cy",
            "gas turb-combined cy",
            "gas turb. & heat rec",
            "gas turb/comb cycle",
            "gas turb/comb. cyc",
            "gas turb/cumb. cyc",
            "gas turb/cumbus cycl",
            "gas turbine / steam",
            "gas turbine-combined cycle",
            "gas turbine/ steam",
            "gas turbine/steam",
            "gas/steam",
            "gas/steam comb cycle",
            "gas/steam comb. cycl",
            "gasturb/comb cycle",
            "igcc",
            "igcc/gas turbine",
            "ngcc",
            "st/gas turb comb cyc",
            "steam & cc",
            "steam & comb cycle",
            "steam & combined ce",
            "steam & gas turbine",
            "steam (comb. cycle)",
            "steam (comb. cycle)gas turbine/steam",
            "steam and cc",
            "steam cc",
            "steam comb cycle",
            "steam/gas turb.",
            "steam/gas turbine",
        },
        "nuclear": {
            "nuclear",
            "nuclear (3)",
            "steam(nuclear)",
            "nuclear(see note4)" "nuclear steam",
            "nuclear turbine",
            "nuclear - steam",
            "nuclear (a)(b)(c)",
            "nuclear (b)(c)",
            "* nuclear",
            "nuclear (b) (c)",
            "nuclear (see notes)",
            "steam (nuclear)",
            "* nuclear (note 2)",
            "nuclear (note 2)",
            "nuclear (see note 2)",
            "nuclear(see note4)",
            "nuclear steam",
            "nuclear(see notes)",
            "nuclear-steam",
            "nuclear (see note 3)",
        },
        "geothermal": {"steam - geothermal", "steam_geothermal", "geothermal"},
        "internal_combustion": {
            "*int combustion (1)",
            "*int. combustion (1)",
            "*internal combust'n",
            "auke bay internal combustion",
            "auke bay internal combustion:",
            "available for specific diesel unit.",
            "available for specific diesel units.",
            "a-internal combustion unit",
            "black start diesel units",
            "black start diesel units:",
            "comb. cyc",
            "comb. turbine",
            "combustion",
            "conemaugh - internal combustion",
            "diesal",
            "diesel",
            "diesel engine",
            "diesel generators",
            "diesel plant",
            "diesel plants",
            "diesel plants:",
            "diesel:",
            "diesel turbine",
            "east hampton diesel",
            "gold creek internal combustion:",
            "ic",
            "int combust (note 1)",
            "int combust (note1)",
            "int combust - note 1",
            "int. combust (note1)",
            "int. combust - note1",
            "int. combustine",
            "int. combustion",
            "int. combustion (1)",
            "int.combustine",
            "interal",
            "internal conbustion",
            "interal combustion:",
            "internal combustion:",
            "internal",
            "internal comb",
            "internal comb recip",
            "internal comb.",
            "internal comb.",
            "internal combust.",
            "internal combustion",
            "internal combustion :",
            "internal combustion auxiliary",
            "internal combustion peaking units",
            "internal combustion plants:",
            "internal combustion reciprocating",
            "internal combustion-diesel",
            "internal combustion:",
            "internal combustion: (emergency standby",
            "internal conbustion:",
            "internal_combustion",
            "international combustion",
            "internl combustion",
            "internl combustion",
            "lemon creek internal combustion",
            "lemon creek internal combustion:",
            "montauk diesel",
            "other / internal combustion :",
            "other/internal combustion",
            "reciprocating engine",
            "shawville - internal combustion",
            "steam internal comb",
        },
        "wind": {
            "(7) montfort wind farm was purchased dec. 2012.",
            "rosiere wind turbines",
            "suisun wind power plant (r)",
            "(7) montfort wind farm was purchased dec. 2012.",
            "wind",
            "wind energy",
            "wind turbine",
            "wind - turbine",
            "wind generation",
            "wind turbin",
            "wind turbine plant",
            "wind turbine plants",
            "wind turbine plants:",
            "wind turbine:",
            "wind turbines",
            "wind:",
        },
        "photovoltaic": {
            "community solar gardens",
            "other production - solar",
            "photovoltaic",
            "sixth street solar",
            "solar",
            "solar array",
            "solar electric",
            "solar photovoltaic",
            "solar photovoltiac",
            "solar plant",
            "solar plants",
            "solar plants:",
            "solar project",
            "solar:",
            "proj dvlpmnt div solar photovoltaic project:",
            "photo voltaic generating plants:",
        },
        "solar_thermal": {"solar thermal", "solar_thermal"},
        "fuel_cell": {
            "fuel cell",
            "fuel_cell",
            "fuel cell:",
        },
        "hydro": {
            "(1) the dixon hydro electric generating facility",
            "(6) excludes portion allocated to other water",
            "(hydroelectric facility)",
            "a - hydro license project no. 2315.",
            "a-hydro license project no. 2315",
            "additional hydro plant",
            "all other hydro misc. plant, resv., etc.",
            "all other hydro miscellaneous plant",
            "all other hydro plant, resv., misc., etc.",
            "conventional hydro",
            "hydraulic",
            "hydraulic (1):",
            "hydraulic race (project no. 2424)(note 1)",
            "hydraulic:",
            "hydro",
            "hydro -",
            "hydro : licensed proj. no.",
            "hydro electric",
            "hydro lic project no.",
            "hydro lic. project no.",
            "hydro license",
            "hydro license no.",
            "hydro plant:",
            "hydro plants",
            "hydro plants:",
            "hydro plants: licensed proj no.",
            "hydro plants: licensed proj. no.",
            "hydro-electric",
            "hydro-neal shoals",
            "hydro:",
            "hydro: lic project no.",
            "hydro: lic. project no.",
            "hydro: license no.",
            "hydroelectric",
            "hydroelectric : licensed proj. no.",
            "hydroelectric generating plants",
            "hydroelectric generating plants :",
            "hydroelectric generating plants:",
            "hydroelectric: licensed proj. no.",
            "hyrdo",
            "neal shoals-hydro license",
            "other hydro assets (resvr., strm gages, etc.)",
            "other hydro assets (resvr.,strm gages, etc.)",
            "other hydro costs:",
            "research and development costs (hydros)",
            "small hydroelectric generating plants:",
            "small hydroelectric plants:",
            "talcville hydro station(project no. 4402)(note 1",
            "the hydro plant above",
            "unallocated - hydros",
            "unallocated hydros",
        },
        "storage": {
            "daily storage",
            "storage",
            "storage (re-reg)",
            "storage - dailey",
            "storage - daily",
            "storage - weekly",
            "storage reservoir",
            "weekly storage",
        },
        "na_category": {
            "",
            "(left blank)",
            "(see note 5)",
            "(see note 7)",
            "--",
            "0",
            "0.0000",
            "combined heat/power",
            "common",
            "common expenses",
            "conventional",
            "expenses commom to",
            "expenses common to",
            "footnote",
            "fuel facilities",
            "fuel oil",
            "gas",
            "n/a",
            "na",
            "na_category",
            "non-applicable",
            "none",
            "none.",
            "not applicable",
            "not applicable.",
            "not in service",
            "not in service in",
            "oil",
            "other",
            "other generation",
            "other generation plt",
            "package",
            "peach bottom",
            "plant retired- 2013",
            "plants sold in 1999",
            "resp share (note 11)",
            "resp share - note 11",
            "resp share - note 5",
            "resp. share (note 2)",
            "resp. share (note 3)",
            "resp. share (note 4)",
            "resp. share (note 6)",
            "resp. share (note 8)",
            "resp. share (note 9)",
            "resp. share (note11)",
            "resp. share (note2)",
            "resp. share (note4)",
            "resp. share (note6)",
            "resp. share (note8)",
            "resp. share - note 2",
            "resp. share - note 4",
            "resp. share - note 6",
            "resp. share - note 7",
            "resp. share - note 8",
            "resp. share - note 9",
            "resp. share note 11",
            "respondent's share",
            "retired",
            "retired - 2013",
            "retired in 1978",
            "retired in 1981",
            "see pgs 402.1-402.3",
            "see pgs 403.1-403.9",
            "sold april 16, 1999",
            "sold may 07, 1999",
            "storage/pipelines",
            "unit total (note 2)",
            "unit total (note 3)",
            "unit total (note2)",
            "unit total (note3)",
            "waste heat",
            "—",
        },
    }
}
"""
A mapping from canonical plant kinds (keys) to the associated freeform strings (values)
identified as being associated with that kind of plant in the FERC Form 1 raw data.
There are many strings that weren't categorized, Solar and Solar Project were not
classified as these do not indicate if they are solar thermal or photovoltaic. Variants
on Steam (e.g. "steam 72" and "steam and gas") were classified based on additional
research of the plants on the Internet.
"""

PLANT_TYPE_CATEGORIES_HYDRO: dict[str, set[str]] = {
    "categories": {
        "hydro": {
            "conventional",
            "hydro",
            "hydro -",
            "hydro : licensed proj. no.",
            "hydro electric",
            "hydro lic project no.",
            "hydro license no.",
            "hydro plant:",
            "hydro plants",
            "hydro plants:",
            "hydro plants: licensed proj no.",
            "hydro plants: licensed proj. no.",
            "hydro-electric",
            "hydro:",
            "hydro: lic project no.",
            "hydro: lic. project no.",
            "hydro: license no.",
            "hydroelectric",
            "hydroelectric : licensed proj. no.",
            "hydroelectric generating plants",
            "hydroelectric generating plants :",
            "hydroelectric generating plants:",
            "hydroelectric: licensed proj. no.",
            "hyrdo",
            "small hydroelectric plants:",
            "unallocated hydros",
        },
        "run_of_river": {
            "run_of_river",
            "run ofriver",
            "r of r",
            "run - of - river",
            "run -of - river",
            "run of river",
            "run ofriver",
            "run-of -river",
            "run-of river",
            "run-of-river",
        },
        "run_of_river_with_storage": {
            "run_of_river_with_storage",
            "run-of-river/storage",
            "r of r/ storage",
            "r of r/storage",
            "r or r/storage",
            "run of river-stge",
            "run of river-stora",
            "run of river-storage",
            "run-of river;storage",
            "run-of-river/stora",
            "run-of-river/storage",
            "run-of-river; sto",
            "run-of-river; stor",
            "run-of-river; stora",
            "run-of-river; storag",
            "run-of-river;stor",
            "run-of-river;stora",
            "run-of-river;storage",
        },
        "storage": {
            "daily storage",
            "storage",
            "storage (re-reg)",
            "storage - dailey",
            "storage - daily",
            "storage - weekly",
            "storage reservoir",
            "weekly storage",
        },
        "na_category": {
            "na_category",
            "(see note 7, page",
            "-",
            "---not applicable---",
            "--not applicable--",
            "00",
            "blank page",
            "gap common",
            "not used",
            "peaking",
            "schedule is n/a",
            "see footnote",
            "see note 7,p.403fn",
            "see note 7,pg 4031",
            "sold in 1999",
            "",
            "not applicable.",
            "gas turbine",
            "na",
            "not applicable",
            "0",
            "none",
            "none.",
            "n/a",
        },
    }
}
"""
A mapping from canonical plant kinds (keys) to the associated freeform strings (values)
identified as being associated with that kind of plant in the FERC Form 1 Hydro Plants
data. These are seperated out from the rest of the plant types due to the difference in
languaged used to refer to hydro vs. other types of plants. For example: "conventional"
in the context of a hydro plant means that it is conventional hydro-electric. In the
context of the steam table, however, it's unclear what conventional means.
"""


CONSTRUCTION_TYPE_CATEGORIES: dict[str, set[str]] = {
    "categories": {
        "outdoor": {
            "outdoor",
            "outdoor boiler",
            "full outdoor",
            "outdoor boiler",
            "outdoor boilers",
            "outboilers",
            "fuel outdoor",
            "full outdoor",
            "outdoors",
            "outdoor",
            "boiler outdoor& full",
            "boiler outdoor&full",
            "outdoor boiler& full",
            "full -outdoor",
            "outdoor steam",
            "outdoor boiler",
            "ob",
            "outdoor automatic",
            "outdoor repower",
            "full outdoor boiler",
            "fo",
            "outdoor boiler & ful",
            "full-outdoor",
            "fuel outdoor",
            "outoor",
            "outdoor",
            "outdoor  boiler&full",
            "boiler outdoor &full",
            "outdoor boiler &full",
            "boiler outdoor & ful",
            "outdoor-boiler",
            "outdoor - boiler",
            "outdoor const.",
            "4 outdoor boilers",
            "3 outdoor boilers",
            "full outdoor",
            "full outdoors",
            "full oudoors",
            "outdoor (auto oper)",
            "outside boiler",
            "outdoor boiler&full",
            "outdoor hrsg",
            "outdoor hrsg",
            "outdoor-steel encl.",
            "boiler-outdr & full",
            "con.& full outdoor",
            "partial outdoor",
            "outdoor (auto. oper)",
            "outdoor (auto.oper)",
            "outdoor construction",
            "1 outdoor boiler",
            "2 outdoor boilers",
            "outdoor enclosure",
            "2 outoor boilers",
            "boiler outdr.& full",
            "boiler outdr. & full",
            "ful outdoor",
            "outdoor-steel enclos",
            "outdoor (auto oper.)",
            "con. & full outdoor",
            "outdore",
            "boiler & full outdor",
            "full & outdr boilers",
            "outodoor (auto oper)",
            "outdoor steel encl.",
            "full outoor",
            "boiler & outdoor ful",
            "otdr. blr. & f. otdr",
            "f.otdr & otdr.blr.",
            "oudoor (auto oper)",
            "outdoor constructin",
            "f. otdr. & otdr. blr",
            "outdoor boiler & fue",
            "outdoor boiler &fuel",
            "conventional; outdoor",
            "coven./outdoor",
        },
        "semioutdoor": {
            "semioutdoor",
            "more than 50% outdoors",
            "more than 50% outdoo",
            "more than 50% outdos",
            "over 50% outdoor",
            "over 50% outdoors",
            "semi-outdoor",
            "semi - outdoor",
            "semi outdoor",
            "semi-enclosed",
            "semi-outdoor boiler",
            "semi outdoor boiler",
            "semi- outdoor",
            "semi - outdoors",
            "semi -outdoor" "conven & semi-outdr",
            "conv & semi-outdoor",
            "conv & semi- outdoor",
            "convent. semi-outdr",
            "conv. semi outdoor",
            "conv(u1)/semiod(u2)",
            "conv u1/semi-od u2",
            "conv-one blr-semi-od",
            "convent semioutdoor",
            "conv. u1/semi-od u2",
            "conv - 1 blr semi od",
            "conv. ui/semi-od u2",
            "conv-1 blr semi-od",
            "conven. semi-outdoor",
            "conv semi-outdoor",
            "u1-conv./u2-semi-od",
            "u1-conv./u2-semi -od",
            "convent. semi-outdoo",
            "u1-conv. / u2-semi",
            "conven & semi-outdr",
            "semi -outdoor",
            "outdr & conventnl",
            "conven. full outdoor",
            "conv. & outdoor blr",
            "conv. & outdoor blr.",
            "conv. & outdoor boil",
            "conv. & outdr boiler",
            "conv. & out. boiler",
            "convntl,outdoor blr",
            "outdoor & conv.",
            "2 conv., 1 out. boil",
            "outdoor/conventional",
            "conv. boiler outdoor",
            "conv-one boiler-outd",
            "conventional outdoor",
            "conventional outdor",
            "conv. outdoor boiler",
            "conv.outdoor boiler",
            "conventional outdr.",
            "conven,outdoorboiler",
            "conven full outdoor",
            "conven,full outdoor",
            "1 out boil, 2 conv",
            "conv. & full outdoor",
            "conv. & outdr. boilr",
            "conv outdoor boiler",
            "convention. outdoor",
            "conv. sem. outdoor",
            "convntl, outdoor blr",
            "conv & outdoor boil",
            "conv & outdoor boil.",
            "outdoor & conv",
            "conv. broiler outdor",
            "1 out boilr, 2 conv",
            "conv.& outdoor boil.",
            "conven,outdr.boiler",
            "conven,outdr boiler",
            "outdoor & conventil",
            "1 out boilr 2 conv",
            "conv & outdr. boilr",
            "conven, full outdoor",
            "conven full outdr.",
            "conven, full outdr.",
            "conv/outdoor boiler",
            "convnt'l outdr boilr",
            "1 out boil 2 conv",
            "conv full outdoor",
            "conven, outdr boiler",
            "conventional/outdoor",
            "conv&outdoor boiler",
            "outdoor & convention",
            "conv & outdoor boilr",
            "conv & full outdoor",
            "convntl. outdoor blr",
            "conv - ob",
            "1conv'l/2odboilers",
            "2conv'l/1odboiler",
            "conv-ob",
            "conv.-ob",
            "1 conv/ 2odboilers",
            "2 conv /1 odboilers",
            "conv- ob",
            "conv -ob",
            "con sem outdoor",
            "cnvntl, outdr, boilr",
            "less than 50% outdoo",
            "less than 50% outdoors",
            "under 50% outdoor",
            "under 50% outdoors",
            "1cnvntnl/2odboilers",
            "2cnvntnl1/1odboiler",
            "con & ob",
            "combination (b)",
            "indoor & outdoor",
            "conven. blr. & full",
            "conv. & otdr. blr.",
            "combination",
            "indoor and outdoor",
            "conven boiler & full",
            "2conv'l/10dboiler",
            "4 indor/outdr boiler",
            "4 indr/outdr boilerr",
            "4 indr/outdr boiler",
            "indoor & outdoof",
        },
        "conventional": {
            "conventional",
            "conventional",
            "conventional boiler",
            "conventional - boiler",
            "conv-b",
            "conventionall",
            "convention",
            "conventional",
            "coventional",
            "conven full boiler",
            "c0nventional",
            "conventtional",
            "convential" "underground",
            "conventional bulb",
            "conventrional",
            "*conventional",
            "convential",
            "convetional",
            "conventioanl",
            "conventioinal",
            "conventaional",
            "indoor construction",
            "convenional",
            "conventional steam",
            "conventinal",
            "convntional",
            "conventionl",
            "conventionsl",
            "conventiional",
            "convntl steam plants",
            "indoor const.",
            "full indoor",
            "indoor",
            "indoor automatic",
            "indoor boiler",
            "indoor boiler and steam turbine",
            "(peak load) indoor",
            "conventionl,indoor",
            "conventionl, indoor",
            "conventional, indoor",
            "conventional;outdoor",
            "conven./outdoor",
            "conventional;semi-ou",
            "comb. cycle indoor",
            "comb cycle indoor",
            "3 indoor boiler",
            "2 indoor boilers",
            "1 indoor boiler",
            "2 indoor boiler",
            "3 indoor boilers",
            "fully contained",
            "conv - b",
            "conventional/boiler",
            "cnventional",
            "comb. cycle indooor",
            "sonventional",
            "ind enclosures",
            "conentional",
            "conventional - boilr",
            "indoor boiler and st",
            "conven",
        },
        "na_category": {
            "na_category",
            "na",
            "",
            "automatic operation",
            "comb. turb. installn",
            "comb. turb. instaln",
            "com. turb. installn",
            "n/a",
            "for detailed info.",
            "for detailed info",
            "combined cycle",
            "not applicable",
            "gas",
            "heated individually",
            "metal enclosure",
            "pressurized water",
            "nuclear",
            "jet engine",
            "gas turbine",
            "storage/pipelines",
            "0",
            "during 1994",
            "peaking - automatic",
            "gas turbine/int. cm",
            "2 oil/gas turbines",
            "wind",
            "package",
            "mobile",
            "auto-operated",
            "steam plants",
            "other production",
            "all nuclear plants",
            "other power gen.",
            "automatically operad",
            "automatically operd",
            "circ fluidized bed",
            "jet turbine",
            "gas turbne/int comb",
            "automatically oper.",
            "retired 1/1/95",
            "during 1995",
            "1996. plant sold",
            "reactivated 7/1/96",
            "gas turbine/int comb",
            "portable",
            "head individually",
            "automatic opertion",
            "peaking-automatic",
            "cycle",
            "full order",
            "circ. fluidized bed",
            "gas turbine/intcomb",
            "0.0000",
            "none",
            "2 oil / gas",
            "block & steel",
            "and 2000",
            "comb.turb. instaln",
            "automatic oper.",
            "pakage",
            "---",
            "—",
            "n/a (ct)",
            "comb turb instain",
            "ind encloures",
            "2 oil /gas turbines",
            "combustion turbine",
            "1970",
            "gas/oil turbines",
            "combined cycle steam",
            "pwr",
            "2 oil/ gas",
            "2 oil / gas turbines",
            "gas / oil turbines",
            "no boiler",
            "internal combustion",
            "gasturbine no boiler",
            "boiler",
            "tower -10 unit facy",
            "gas trubine",
            "4 gas/oil trubines",
            "2 oil/ 4 gas/oil tur",
            "5 gas/oil turbines",
            "tower 16",
            "2 on 1 gas turbine",
            "tower 23",
            "tower -10 unit",
            "tower - 101 unit",
            "3 on 1 gas turbine",
            "tower - 10 units",
            "tower - 165 units",
            "wind turbine",
            "fixed tilt pv",
            "tracking pv",
            "o",
            "wind trubine",
            "wind generator",
            "subcritical",
            "sucritical",
            "simple cycle",
            "simple & reciprocat",
            "solar",
            "pre-fab power plant",
            "prefab power plant",
            "prefab. power plant",
            "pump storage",
            "underground",
            "see page 402",
            "conv. underground",
            "conven. underground",
            "conventional (a)",
            "non-applicable",
            "duct burner",
            "see footnote",
            "simple and reciprocat",
            "--not applicable--",
            "---not applicable---",
            "not applicable.",
            "schedule is n/a",
            "none.",
            "see note",
            "not appicable",
            "pv tracking",
        },
    }
}
"""
A dictionary of construction types (keys) and lists of construction type strings
associated with each type (values) from FERC Form 1.

There are many strings that weren't categorized, including crosses between conventional
and outdoor, PV, wind, combined cycle, and internal combustion. The lists are broken out
into the two types specified in Form 1: conventional and outdoor. These lists are
inclusive so that variants of conventional (e.g. "conventional full") and outdoor (e.g.
"outdoor full" and "outdoor hrsg") are included.
"""

##############################################################################
# Fully assembled set of FERC 1 transformation parameters
##############################################################################
TRANSFORM_PARAMS = {
    "fuel_ferc1": {
        "rename_columns_ferc1": {
            "dbf": {
                "columns": {
                    "respondent_id": "utility_id_ferc1_dbf",
                    "plant_name": "plant_name_ferc1",
                    "fuel": "fuel_type_code_pudl",
                    "fuel_unit": "fuel_units",
                    "fuel_avg_heat": "fuel_btu_per_unit",
                    "fuel_quantity": "fuel_consumed_units",
                    "fuel_cost_burned": "fuel_cost_per_unit_burned",
                    "fuel_cost_delvd": "fuel_cost_per_unit_delivered",
                    # Note: Original fuel_cost_btu is misleadingly named
                    "fuel_cost_btu": "fuel_cost_per_mmbtu",
                    "fuel_generaton": "fuel_btu_per_kwh",
                    "report_prd": "report_prd",
                    "row_prvlg": "row_prvlg",
                    "report_year": "report_year",
                    "row_number": "row_number",
                    "fuel_cost_kwh": "fuel_cost_per_kwh",
                    "spplmnt_num": "spplmnt_num",
                    "row_seq": "row_seq",
                }
            },
            "xbrl": {
                "columns": {
                    "plant_name_axis": "plant_name_ferc1",
                    "fuel_kind_axis": "fuel_type_code_pudl",
                    "fuel_unit": "fuel_units",
                    "fuel_burned_average_heat_content": "fuel_btu_per_unit",
                    "quantity_of_fuel_burned": "fuel_consumed_units",
                    "average_cost_of_fuel_per_unit_burned": "fuel_cost_per_unit_burned",
                    "average_cost_of_fuel_per_unit_as_delivered": "fuel_cost_per_unit_delivered",
                    "average_cost_of_fuel_burned_per_million_british_thermal_unit": "fuel_cost_per_mmbtu",
                    "average_british_thermal_unit_per_kilowatt_hour_net_generation": "fuel_btu_per_kwh",
                    "average_cost_of_fuel_burned_per_kilowatt_hour_net_generation": "fuel_cost_per_kwh",
                    "report_year": "report_year",
                    "fuel_kind": "fuel_kind",
                    "end_date": "end_date",
                    "entity_id": "utility_id_ferc1_xbrl",
                    "start_date": "start_date",
                }
            },
        },
        "categorize_strings": {
            "fuel_type_code_pudl": FUEL_CATEGORIES,
            "fuel_units": FUEL_UNIT_CATEGORIES,
        },
        "convert_units": {
            "fuel_btu_per_unit": BTU_TO_MMBTU,
            "fuel_btu_per_kwh": BTU_PERKWH_TO_MMBTU_PERMWH,
            "fuel_cost_per_kwh": PERKWH_TO_PERMWH,
        },
        "normalize_strings": {
            "plant_name_ferc1": FERC1_STRING_NORM,
            "fuel_type_code_pudl": FERC1_STRING_NORM,
            "fuel_units": FERC1_STRING_NORM,
        },
        "correct_units": [
            COAL_MMBTU_PER_UNIT_CORRECTIONS,
            GAS_MMBTU_PER_UNIT_CORRECTIONS,
            OIL_MMBTU_PER_UNIT_CORRECTIONS,
            COAL_COST_PER_MMBTU_CORRECTIONS,
            GAS_COST_PER_MMBTU_CORRECTIONS,
            OIL_COST_PER_MMBTU_CORRECTIONS,
        ],
        "drop_invalid_rows": [
            {
                "invalid_values": [0, pd.NA, np.nan, ""],
                "required_valid_cols": [
                    "fuel_consumed_units",
                    "fuel_cost_per_mmbtu",
                    "fuel_cost_per_mwh",
                    "fuel_cost_per_unit_delivered",
                    "fuel_cost_per_unit_burned",
                    "fuel_mmbtu_per_unit",
                    "fuel_units",
                ],
            },
            {
                "invalid_values": [
                    pd.NA,
                    np.nan,
                    "",
                    "must 123",
                    "must 456",
                    "ant1-3",
                    "elk 1-3",
                ],
                "required_valid_cols": ["plant_name_ferc1"],
            },
        ],
    },
    "plants_steam_ferc1": {
        "rename_columns_ferc1": {
            "dbf": {
                "columns": {
                    "cost_structure": "capex_structures",
                    "expns_misc_power": "opex_misc_power",
                    "plant_name": "plant_name_ferc1",
                    "plnt_capability": "plant_capability_mw",
                    "expns_plants": "opex_plants",
                    "expns_misc_steam": "opex_misc_steam",
                    "cost_per_kw": "capex_per_kw",
                    "when_not_limited": "not_water_limited_capacity_mw",
                    "asset_retire_cost": "asset_retirement_cost",
                    "expns_steam_othr": "opex_steam_other",
                    "expns_transfer": "opex_transfer",
                    "expns_engnr": "opex_engineering",
                    "avg_num_of_emp": "avg_num_employees",
                    "cost_of_plant_to": "capex_total",
                    "expns_rents": "opex_rents",
                    "tot_prdctn_expns": "opex_production_total",
                    "plant_kind": "plant_type",
                    "respondent_id": "utility_id_ferc1_dbf",
                    "expns_operations": "opex_operations",
                    "cost_equipment": "capex_equipment",
                    "type_const": "construction_type",
                    "plant_hours": "plant_hours_connected_while_generating",
                    "expns_coolants": "opex_coolants",
                    "expns_fuel": "opex_fuel",
                    "when_limited": "water_limited_capacity_mw",
                    "expns_kwh": "opex_per_kwh",
                    "expns_allowances": "opex_allowances",
                    "expns_steam": "opex_steam",
                    "yr_const": "construction_year",
                    "yr_installed": "installation_year",
                    "expns_boiler": "opex_boiler",
                    "peak_demand": "peak_demand_mw",
                    "cost_land": "capex_land",
                    "tot_capacity": "capacity_mw",
                    "net_generation": "net_generation_kwh",
                    "expns_electric": "opex_electric",
                    "expns_structures": "opex_structures",
                    "report_year": "report_year",
                    "report_prd": "report_prd",
                    "row_prvlg": "row_prvlg",
                    "row_number": "row_number",
                    "spplmnt_num": "spplmnt_num",
                    "row_seq": "row_seq",
                },
            },
            "xbrl": {
                "columns": {
                    "cost_of_structures_and_improvements_steam_production": "capex_structures",
                    "miscellaneous_steam_power_expenses": "opex_misc_power",
                    "plant_name_axis": "plant_name_ferc1",
                    "net_continuous_plant_capability": "plant_capability_mw",
                    "maintenance_of_electric_plant_steam_power_generation": "opex_plants",
                    "maintenance_of_miscellaneous_steam_plant": "opex_misc_steam",
                    "cost_per_kilowatt_of_installed_capacity": "capex_per_kw",
                    "net_continuous_plant_capability_not_limited_by_condenser_water": "not_water_limited_capacity_mw",
                    "asset_retirement_costs_steam_production": "asset_retirement_cost",
                    "steam_from_other_sources": "opex_steam_other",
                    "steam_transferred_credit": "opex_transfer",
                    "maintenance_supervision_and_engineering_steam_power_generation": "opex_engineering",
                    "plant_average_number_of_employees": "avg_num_employees",
                    "cost_of_plant": "capex_total",
                    "rents_steam_power_generation": "opex_rents",
                    "power_production_expenses_steam_power": "opex_production_total",
                    "plant_kind": "plant_type",
                    "operation_supervision_and_engineering_expense": "opex_operations",
                    "cost_of_equipment_steam_production": "capex_equipment",
                    "plant_construction_type": "construction_type",
                    "plant_hours_connected_to_load": "plant_hours_connected_while_generating",
                    "coolants_and_water": "opex_coolants",
                    "fuel_steam_power_generation": "opex_fuel",
                    "net_continuous_plant_capability_limited_by_condenser_water": "water_limited_capacity_mw",
                    "expenses_per_net_kilowatt_hour": "opex_per_kwh",
                    "allowances": "opex_allowances",
                    "steam_expenses_steam_power_generation": "opex_steam",
                    "year_plant_originally_constructed": "construction_year",
                    "year_last_unit_of_plant_installed": "installation_year",
                    "maintenance_of_boiler_plant_steam_power_generation": "opex_boiler",
                    "net_peak_demand_on_plant": "peak_demand_mw",
                    "cost_of_land_and_land_rights_steam_production": "capex_land",
                    "installed_capacity_of_plant": "capacity_mw",
                    "net_generation_excluding_plant_use": "net_generation_kwh",
                    "electric_expenses_steam_power_generation": "opex_electric",
                    "maintenance_of_structures_steam_power_generation": "opex_structures",
                    "report_year": "report_year",
                    "order_number": "order_number",
                    "plant_name": "plant_name",
                    "entity_id": "utility_id_ferc1_xbrl",
                    "end_date": "end_date",
                    "start_date": "start_date",
                    "date": "date",
                },
            },
        },
        "normalize_strings": {
            "plant_name_ferc1": FERC1_STRING_NORM,
            "construction_type": FERC1_STRING_NORM,
            "plant_type": FERC1_STRING_NORM,
        },
        "nullify_outliers": {
            "construction_year": VALID_PLANT_YEARS,
            "installation_year": VALID_PLANT_YEARS,
        },
        "categorize_strings": {
            "construction_type": CONSTRUCTION_TYPE_CATEGORIES,
            "plant_type": PLANT_TYPE_CATEGORIES,
        },
        "convert_units": {
            "capex_per_kw": PERKW_TO_PERMW,
            "opex_per_kwh": PERKWH_TO_PERMWH,
            "net_generation_kwh": KWH_TO_MWH,
        },
        "drop_invalid_rows": [
            {
                "invalid_values": [0, "0", pd.NA, np.nan, "", "none"],
                "allowed_invalid_cols": [
                    "record_id",
                    "utility_id_ferc1",
                    "utility_id_ferc1_dbf",
                    "utility_id_ferc1_xbrl",
                    "plant_name_ferc1",
                    "report_year",
                    "date",
                    "start_date",
                    "end_date",
                    "order_number",
                    "plant_name",
                    "plant_type",
                ],
            },
            {
                "invalid_values": [
                    pd.NA,
                    np.nan,
                    "",
                ],
                "required_valid_cols": ["plant_name_ferc1"],
            },
        ],
    },
    "plants_hydro_ferc1": {
        "rename_columns_ferc1": {
            "dbf": {
                "columns": {
                    "plant_name": "plant_name_ferc1",
                    "project_no": "project_num",
                    "yr_const": "construction_year",
                    "plant_kind": "plant_type",
                    "plant_const": "construction_type",
                    "yr_installed": "installation_year",
                    "tot_capacity": "capacity_mw",
                    "peak_demand": "peak_demand_mw",
                    "plant_hours": "plant_hours_connected_while_generating",
                    "favorable_cond": "net_capacity_favorable_conditions_mw",
                    "adverse_cond": "net_capacity_adverse_conditions_mw",
                    "avg_num_of_emp": "avg_num_employees",
                    "net_generation": "net_generation_kwh",
                    "cost_of_land": "capex_land",
                    "cost_structure": "capex_structures",
                    "cost_facilities": "capex_facilities",
                    "cost_equipment": "capex_equipment",
                    "cost_roads": "capex_roads",
                    "cost_plant_total": "capex_total",
                    "cost_per_kw": "cost_per_kw",
                    "expns_operations": "opex_operations",
                    "expns_water_pwr": "opex_water_for_power",
                    "expns_hydraulic": "opex_hydraulic",
                    "expns_electric": "opex_electric",
                    "expns_generation": "opex_generation_misc",
                    "expns_rents": "opex_rents",
                    "expns_structures": "opex_structures",
                    "expns_dams": "opex_dams",
                    "expns_plant": "opex_plant",
                    "expns_misc_plant": "opex_misc_plant",
                    "expns_kwh": "opex_per_kwh",
                    "expns_engnr": "opex_engineering",
                    "expns_total": "opex_total",
                    "asset_retire_cost": "asset_retirement_cost",
                    "report_year": "report_year",
                    "respondent_id": "utility_id_ferc1_dbf",
                    "report_prd": "report_prd",
                    "row_prvlg": "row_prvlg",
                    "row_number": "row_number",
                    "row_seq": "row_seq",
                    "spplmnt_num": "spplmnt_num",
                },
            },
            "xbrl": {
                "columns": {
                    "entity_id": "utility_id_ferc1_xbrl",
                    "plant_name": "plant_name_ferc1",
                    "order_number": "order_number",
                    "report_year": "report_year",
                    "date": "date",
                    "start_date": "start_date",
                    "end_date": "end_date",
                    "licensed_project_number_and_hydroelectric_plant_name_axis": "project_num_and_plant_name_ferc1",
                    "installed_capacity_of_plant": "capacity_mw",
                    "cost_of_plant": "capex_total",
                    "equipment_costs_hydroelectric_production": "capex_equipment",
                    "cost_of_reservoirs_dams_and_waterways_hydroelectric_production": "capex_facilities",
                    "cost_of_structures_and_improvements_hydroelectric_production": "capex_structures",  # should we rename this as capex_structures_and_improvements?
                    "asset_retirement_costs_hydroelectric_production": "asset_retirement_cost",  # should we rename this to capex_asset_retirement?
                    "cost_of_land_and_land_rights_hydroelectric_production": "capex_land",
                    "cost_of_roads_railroads_and_bridges_hydroelectric_production": "capex_roads",
                    "licensed_project_number": "project_num",
                    "net_plant_capability_under_most_favorable_operating_conditions": "net_capacity_favorable_conditions_mw",
                    "plant_construction_type": "construction_type",
                    "net_generation_excluding_plant_use": "net_generation_kwh",
                    "expenses_per_net_kilowatt_hour": "opex_per_kwh",
                    "cost_per_kilowatt_of_installed_capacity": "capex_per_kw",
                    "maintenance_of_structures_hydraulic_power_generation": "opex_structures",
                    "maintenance_supervision_and_engineering_hydraulic_power_generation": "opex_engineering",
                    "maintenance_of_miscellaneous_hydraulic_plant": "opex_misc_plant",
                    "hydraulic_expenses": "opex_hydraulic",
                    "water_for_power": "opex_water_for_power",
                    "electric_expenses_hydraulic_power_generation": "opex_electric",
                    "maintenance_of_electric_plant_hydraulic_power_generation": "opex_plant",
                    "plant_hours_connected_to_load": "plant_hours_connected_while_generating",
                    "power_production_expenses_hydraulic_power": "opex_total",
                    "miscellaneous_hydraulic_power_generation_expenses": "opex_generation_misc",
                    "net_peak_demand_on_plant": "peak_demand_mw",
                    "year_last_unit_of_plant_installed": "installation_year",
                    "plant_average_number_of_employees": "avg_num_employees",
                    "plant_kind": "plant_type",
                    "year_plant_originally_constructed": "construction_year",
                    "maintenance_of_reservoirs_dams_and_waterways": "opex_dams",
                    "rents_hydraulic_power_generation": "opex_rents",
                    "net_plant_capability_under_most_adverse_operating_conditions": "net_capacity_adverse_conditions_mw",
                    "operation_supervision_and_engineering_expense": "opex_operations",
                },
            },
        },
        "normalize_strings": {
            "plant_name_ferc1": FERC1_STRING_NORM,
            "construction_type": FERC1_STRING_NORM,
            "plant_type": FERC1_STRING_NORM,
        },
        "nullify_outliers": {
            "construction_year": VALID_PLANT_YEARS,
            "installation_year": VALID_PLANT_YEARS,
        },
        "categorize_strings": {
            "construction_type": CONSTRUCTION_TYPE_CATEGORIES,
            "plant_type": PLANT_TYPE_CATEGORIES_HYDRO,
        },
        "convert_units": {
            "capex_per_kw": PERKW_TO_PERMW,
            "opex_per_kwh": PERKWH_TO_PERMWH,
            "net_generation_kwh": KWH_TO_MWH,
        },
        "strip_non_numeric_values": {"project_num": {"strip_non_numeric_values": True}},
        "drop_invalid_rows": [
            {
                "invalid_values": [0, "0", pd.NA, np.nan, "", "none", "—"],
                "allowed_invalid_cols": [
                    "record_id",
                    "utility_id_ferc1",
                    "utility_id_ferc1_dbf",
                    "utility_id_ferc1_xbrl",
                    "plant_name_ferc1",
                    "project_num_and_plant_name_ferc1",
                    "report_year",
                    "date",
                    "start_date",
                    "end_date",
                    "order_number",
                    "project_num",
                    "plant_type",
                    "construction_type",
                ],
            },
            {
                "invalid_values": ["0", "—", ""],
                "required_valid_cols": ["plant_name_ferc1"],
            },
        ],
    },
    "plants_small_ferc1": {
        "rename_columns_ferc1": {
            "dbf": {
                "columns": {
                    "plant_name": "plant_name_ferc1",
                    "yr_constructed": "construction_year",
                    "capacity_rating": "capacity_mw",
                    "net_demand": "peak_demand_mw",
                    "net_generation": "net_generation_kwh",
                    "plant_cost": "capex_total",
                    "plant_cost_mw": "capex_per_mw",
                    "operation": "opex_operations",
                    "expns_fuel": "opex_fuel",
                    "expns_maint": "opex_maintenance",
                    "kind_of_fuel": "fuel_type",
                    "fuel_cost": "fuel_cost_per_mmbtu",
                    "report_year": "report_year",
                    "respondent_id": "utility_id_ferc1_dbf",
                    "report_prd": "report_prd",
                    "row_prvlg": "row_prvlg",
                    "row_number": "row_number",
                    "row_seq": "row_seq",
                    "spplmnt_num": "spplmnt_num",
                }
            },
            "xbrl": {
                "columns": {
                    "report_year": "report_year",
                    "entity_id": "utility_id_ferc1_xbrl",
                    "plant_name": "plant_name_ferc1",
                    "start_date": "start_date",
                    "end_date": "end_date",
                    "date": "date",
                    "generation_type": "plant_type",
                    "fuel_cost_per_mmbtus": "fuel_cost_per_mmbtu",
                    "fuel_kind": "fuel_type",
                    "year_plant_originally_constructed": "construction_year",
                    "maintenance_production_expenses": "opex_maintenance",
                    "operating_expenses_excluding_fuel": "opex_operations",
                    "fuel_production_expenses": "opex_fuel",
                    "order_number": "order_number",
                    "net_generation_excluding_plant_use": "net_generation_kwh",
                    "net_peak_demand_on_plant": "peak_demand_mw",
                    "plant_cost_per_mw": "capex_per_mw",
                    "installed_capacity_of_plant": "capacity_mw",
                    "cost_of_plant": "capex_total",
                    "generating_plant_statistics_axis": "generating_plant_statistics",
                },
            },
        },
        "normalize_strings": {
            "plant_name_ferc1": FERC1_STRING_NORM,
            "fuel_type": FERC1_STRING_NORM,
            "plant_type": FERC1_STRING_NORM,
        },
        "nullify_outliers": {
            "construction_year": VALID_PLANT_YEARS,
        },
        "convert_units": {
            "fuel_cost_per_mmbtu": CENTS_PERMMBTU_TO_USD_PERMMBTU,
            "net_generation_kwh": KWH_TO_MWH,
        },
        "categorize_strings": {
            "fuel_type": FUEL_CATEGORIES,
            "fuel_type_from_header": FUEL_CATEGORIES,
            "plant_type": PLANT_TYPE_CATEGORIES,
            "plant_type_from_header": PLANT_TYPE_CATEGORIES,
        },
        "drop_invalid_rows": [
            {"invalid_values": ["header", "note"], "required_valid_cols": ["row_type"]},
            {
                "invalid_values": [0, "0", pd.NA, np.nan, "", "none"],
                "allowed_invalid_cols": [
                    "record_id",
                    "utility_id_ferc1",
                    "utility_id_ferc1_dbf",
                    "utility_id_ferc1_xbrl",
                    "plant_name_ferc1",
                    "generating_plant_statistics",
                    "report_year",
                    "construction_year",
                    "date",
                    "start_date",
                    "end_date",
                    "order_number",
                    "plant_type",
                    "fuel_type",
                    "license_id_ferc1",
                    "notes",
                    "row_type",
                ],
            },
            {"invalid_values": ["0", ""], "required_valid_cols": ["plant_name_ferc1"]},
        ],
    },
    "plant_in_service_ferc1": {
        "rename_columns_ferc1": {
            "xbrl": {
                "columns": {
                    "entity_id": "utility_id_ferc1_xbrl",
                    "report_year": "report_year",
                    "ferc_account_label": "ferc_account_label",
                    "starting_balance": "starting_balance",
                    "additions": "additions",
                    "retirements": "retirements",
                    "adjustments": "adjustments",
                    "transfers": "transfers",
                    "ending_balance": "ending_balance",
                }
            },
            "dbf": {
                "columns": {
                    "report_year": "report_year",
                    "respondent_id": "utility_id_ferc1_dbf",
                    "report_prd": "report_prd",
                    "row_prvlg": "row_prvlg",
                    "row_number": "row_number",
                    "row_seq": "row_seq",
                    "spplmnt_num": "spplmnt_num",
                    "begin_yr_bal": "starting_balance",
                    "addition": "additions",
                    "retirements": "retirements",
                    "adjustments": "adjustments",
                    "transfers": "transfers",
                    "yr_end_bal": "ending_balance",
                }
            },
        },
        "rename_columns_instant_xbrl": {
            "columns": {
                "communication_equipment": "communication_equipment_general_plant",
                "office_furniture_and_equipment": "office_furniture_and_equipment_general_plant",
                "structures_and_improvements": "structures_and_improvements_general_plant",
                "tools_shop_and_garage_equipment": "tools_shop_and_garage_equipment_general_plant",
                "miscellaneous_equipment": "miscellaneous_equipment_general_plant",
                "transportation_equipment": "transportation_equipment_general_plant",
                "laboratory_equipment": "laboratory_equipment_general_plant",
                "land_and_land_rights": "land_and_land_rights_general_plant",
                "stores_equipment": "stores_equipment_general_plant",
                "power_operated_equipment": "power_operated_equipment_general_plant",
                "other_tangible_property": "other_tangible_property_general_plant",
                "installations_on_customers_premises_distribution_plant": "installations_on_customer_premises_distribution_plant",
                "fuel_holders_producers_and_accessories_other_production": "fuel_holders_products_and_accessories_other_production",
                "structures_and_improvement_nuclear_production": "structures_and_improvements_nuclear_production",
                "leased_property_on_customers_premises_distribution_plant": "leased_property_on_customer_premises_distribution_plant",
            }
        },
        "rename_columns_duration_xbrl": {
            "columns": {
                "energy_storage_equipment_production_other_production_additions": "energy_storage_equipment_other_production_additions",
                "energy_storage_equipment_production_other_production_adjustments": "energy_storage_equipment_other_production_adjustments",
                "energy_storage_equipment_production_other_production_retirements": "energy_storage_equipment_other_production_retirements",
                "energy_storage_equipment_production_other_production_transfers": "energy_storage_equipment_other_production_transfers",
                "asset_retirement_costs_for_regional_transmission_and_market_operations_regional_transmission_and_market_operation_plant_additions": "asset_retirement_costs_for_regional_transmission_and_market_operation_plant_regional_transmission_and_market_operation_plant_additions",
                "asset_retirement_costs_for_regional_transmission_and_market_operations_regional_transmission_and_market_operation_plant_adjustments": "asset_retirement_costs_for_regional_transmission_and_market_operation_plant_regional_transmission_and_market_operation_plant_adjustments",
                "asset_retirement_costs_for_regional_transmission_and_market_operations_regional_transmission_and_market_operation_plant_retirements": "asset_retirement_costs_for_regional_transmission_and_market_operation_plant_regional_transmission_and_market_operation_plant_retirements",
                "asset_retirement_costs_for_regional_transmission_and_market_operations_regional_transmission_and_market_operation_plant_transfers": "asset_retirement_costs_for_regional_transmission_and_market_operation_plant_regional_transmission_and_market_operation_plant_transfers",
            }
        },
    },
    "plants_pumped_storage_ferc1": {
        "rename_columns_ferc1": {
            "dbf": {
                "columns": {
                    "report_year": "report_year",
                    "respondent_id": "utility_id_ferc1_dbf",
                    "report_prd": "report_prd",
                    "row_prvlg": "row_prvlg",
                    "row_number": "row_number",
                    "row_seq": "row_seq",
                    "spplmnt_num": "spplmnt_num",
                    "plant_name": "plant_name_ferc1",
                    "tot_capacity": "capacity_mw",
                    "project_no": "project_num",
                    "plant_kind": "construction_type",
                    "peak_demand": "peak_demand_mw",
                    "yr_const": "construction_year",
                    "yr_installed": "installation_year",
                    "plant_hours": "plant_hours_connected_while_generating",
                    "plant_capability": "plant_capability_mw",
                    "avg_num_of_emp": "avg_num_employees",
                    "cost_wheels": "capex_wheels_turbines_generators",
                    "cost_land": "capex_land",
                    "cost_structures": "capex_structures",
                    "cost_facilties": "capex_facilities",
                    "cost_electric": "capex_equipment_electric",
                    "cost_misc_eqpmnt": "capex_equipment_misc",
                    "cost_roads": "capex_roads",
                    "asset_retire_cost": "asset_retirement_cost",
                    "cost_of_plant": "capex_total",
                    "cost_per_kw": "capex_per_kw",
                    "expns_operations": "opex_operations",
                    "expns_water_pwr": "opex_water_for_power",
                    "expns_pump_strg": "opex_pumped_storage",
                    "expns_electric": "opex_electric",
                    "expns_misc_power": "opex_generation_misc",
                    "expns_rents": "opex_rents",
                    "expns_engneering": "opex_engineering",
                    "expns_structures": "opex_structures",
                    "expns_dams": "opex_dams",
                    "expns_plant": "opex_plant",
                    "expns_misc_plnt": "opex_misc_plant",
                    "expns_producton": "opex_production_before_pumping",
                    "expns_kwh": "opex_per_kwh",
                    "pumping_expenses": "opex_pumping",
                    "tot_prdctn_exns": "opex_total",
                    "energy_used": "energy_used_for_pumping_kwh",
                    "net_load": "net_load_kwh",
                    "net_generation": "net_generation_kwh",
                },
            },
            "xbrl": {
                "columns": {
                    "entity_id": "utility_id_ferc1_xbrl",
                    "plant_name": "plant_name_ferc1",
                    "order_number": "order_number",
                    "report_year": "report_year",
                    "date": "date",
                    "start_date": "start_date",
                    "end_date": "end_date",
                    "asset_retirement_costs_pumped_storage_plant": "asset_retirement_cost",
                    "cost_of_accessory_electric_equipment_pumped_storage_plant": "capex_equipment_electric",
                    "cost_of_land_and_land_rights_pumped_storage_plant": "capex_land",
                    "cost_of_miscellaneous_power_plant_equipment_pumped_storage_plant": "capex_equipment_misc",
                    "cost_of_plant": "capex_total",
                    "cost_of_reservoirs_dams_and_waterways_pumped_storage_plant": "capex_facilities",
                    "cost_of_roads_railroads_and_bridges_pumped_storage_plant": "capex_roads",
                    "cost_of_structures_and_improvements_pumped_storage_plant": "capex_structures",
                    "cost_of_water_wheels_turbines_and_generators_pumped_storage_plant": "capex_wheels_turbines_generators",
                    "cost_per_kilowatt_of_installed_capacity": "capex_per_kw",
                    "electric_expenses_pumped_storage_plant": "opex_electric",
                    "energy_used_for_pumping": "energy_used_for_pumping_kwh",
                    "expenses_per_net_kilowatt_hour": "opex_per_kwh",
                    "expenses_per_net_kilowatt_hour_generation_and_pumping": "opex_generation_and_pumping_per_kwh",
                    "installed_capacity_of_plant": "capacity_mw",
                    "licensed_project_number": "project_num",
                    "licensed_project_number_and_pumped_storage_plant_name_axis": "project_num_and_plant_name_ferc1",
                    "maintenance_of_electric_plant_pumped_storage_plant": "opex_plant",
                    "maintenance_of_miscellaneous_pumped_storage_plant": "opex_misc_plant",
                    "maintenance_of_reservoirs_dams_and_waterways_pumped_storage_plant": "opex_dams",
                    "maintenance_of_structures_pumped_storage_plant": "opex_structures",
                    "maintenance_supervision_and_engineering_pumped_storage_plant": "opex_engineering",
                    "miscellaneous_pumped_storage_power_generation_expenses": "opex_generation_misc",
                    "net_continuous_plant_capability": "plant_capability_mw",
                    "net_generation_excluding_plant_use": "net_generation_kwh",
                    "net_output_for_load": "net_load_kwh",
                    "net_peak_demand_on_plant": "peak_demand_mw",
                    "operation_supervision_and_engineering_expense": "opex_operations",
                    "plant_average_number_of_employees": "avg_num_employees",
                    "plant_construction_type": "construction_type",
                    "plant_hours_connected_to_load": "plant_hours_connected_while_generating",
                    "power_production_expense_before_pumping_expenses": "opex_production_before_pumping",
                    "power_production_expenses_pumped_storage_plant": "opex_total",
                    "pumped_storage_expenses": "opex_pumped_storage",
                    "pumping_expenses": "opex_pumping",
                    "rents_pumped_storage_plant": "opex_rents",
                    "water_for_power": "opex_water_for_power",
                    "year_last_unit_of_plant_installed": "installation_year",
                    "year_plant_originally_constructed": "construction_year",
                },
            },
        },
        "normalize_strings": {
            "plant_name_ferc1": FERC1_STRING_NORM,
            "construction_type": FERC1_STRING_NORM,
        },
        "nullify_outliers": {
            "construction_year": VALID_PLANT_YEARS,
            "installation_year": VALID_PLANT_YEARS,
        },
        "categorize_strings": {
            "construction_type": CONSTRUCTION_TYPE_CATEGORIES,
        },
        "convert_units": {
            "capex_per_kw": PERKW_TO_PERMW,
            "opex_per_kwh": PERKWH_TO_PERMWH,
            "opex_generation_and_pumping_per_kwh": PERKWH_TO_PERMWH,
            "net_generation_kwh": KWH_TO_MWH,
            "net_load_kwh": KWH_TO_MWH,
            "energy_used_for_pumping_kwh": KWH_TO_MWH,
        },
        "strip_non_numeric_values": {"project_num": {"strip_non_numeric_values": True}},
        "drop_invalid_rows": [
            {
                "invalid_values": [0, "0", pd.NA, np.nan, "", "none", "—"],
                "allowed_invalid_cols": [
                    "record_id",
                    "utility_id_ferc1",
                    "utility_id_ferc1_dbf",
                    "utility_id_ferc1_xbrl",
                    "plant_name_ferc1",
                    "report_year",
                    "date",
                    "start_date",
                    "end_date",
                    "project_num",
                    "construction_type",
                ],
            },
            {
                "invalid_values": ["0", "—", ""],
                "required_valid_cols": ["plant_name_ferc1"],
            },
        ],
    },
    "purchased_power_ferc1": {
        "rename_columns_ferc1": {
            "dbf": {
                "columns": {
                    "report_prd": "report_prd",
                    "row_number": "row_number",
                    "row_seq": "row_seq",
                    "row_prvlg": "row_prvlg",
                    "spplmnt_num": "spplmnt_num",
                    "report_year": "report_year",
                    "respondent_id": "utility_id_ferc1_dbf",
                    "athrty_co_name": "seller_name",
                    "sttstcl_clssfctn": "purchase_type_code",
                    "rtsched_trffnbr": "tariff",
                    "avgmth_bill_dmnd": "billing_demand_mw",
                    "avgmth_ncp_dmnd": "non_coincident_peak_demand_mw",
                    "avgmth_cp_dmnd": "coincident_peak_demand_mw",
                    "mwh_purchased": "purchased_mwh",
                    "mwh_recv": "received_mwh",
                    "mwh_delvd": "delivered_mwh",
                    "dmnd_charges": "demand_charges",
                    "erg_charges": "energy_charges",
                    "othr_charges": "other_charges",
                    "settlement_tot": "total_settlement",
                }
            },
            "xbrl": {
                "columns": {
                    "end_date": "end_date",
                    "start_date": "start_date",
                    "report_year": "report_year",
                    "order_number": "order_number",
                    "entity_id": "utility_id_ferc1_xbrl",
                    "average_monthly_billing_demand": "billing_demand_mw",
                    "average_monthly_coincident_peak_demand": "coincident_peak_demand_mw",
                    "average_monthly_non_coincident_peak_demand": "non_coincident_peak_demand_mw",
                    "demand_charges_of_purchased_power": "demand_charges",
                    "energy_charges_of_purchased_power": "energy_charges",
                    "energy_delivered_through_power_exchanges": "delivered_mwh",
                    "energy_received_through_power_exchanges": "received_mwh",
                    "megawatt_hours_purchased_for_energy_storage": "purchased_storage_mwh",  # new column
                    "megawatt_hours_purchased_other_than_storage": "purchased_mwh",
                    "name_of_company_or_public_authority_providing_purchased_power": "seller_name",
                    "other_charges_of_purchased_power": "other_charges",
                    "purchased_power_axis": "purchased_power_axis",  # new probably pk col
                    "rate_schedule_tariff_number": "tariff",
                    "settlement_of_power": "total_settlement",
                    "statistical_classification_code": "purchase_type_code",
                }
            },
        },
        "normalize_strings": {
            "seller_name": FERC1_STRING_NORM,
        },
        "strip_non_numeric_values": {
            "billing_demand_mw": {"strip_non_numeric_values": True},
            "non_coincident_peak_demand_mw": {"strip_non_numeric_values": True},
            "coincident_peak_demand_mw": {"strip_non_numeric_values": True},
        },
        "drop_invalid_rows": [
            {
                "invalid_values": [0, "0", pd.NA, np.nan, "", "none", "—"],
                "required_valid_cols": [
                    "purchased_mwh",
                    "received_mwh",
                    "delivered_mwh",
                    "demand_charges",
                    "energy_charges",
                    "other_charges",
                    "total_settlement",
                    "purchased_storage_mwh",
                    "billing_demand_mw",
                    "non_coincident_peak_demand_mw",
                    "coincident_peak_demand_mw",
                ],
            }
        ],
    },
}

"""The full set of parameters used to transform the FERC Form 1 data.

Each item in the dictionary can be used to instantiate a
:class:`pudl.transform.ferc1.Ferc1TableTransformParams` object appropriate for
transforming the table identified by that item's key.
"""
